# Copyright (c) 2024 CenturyArks Co.,Ltd.
#
# These source code may be used (including the use of modified source code) only when using
# CenturyArks's products or services.
#
# Although we have confirmed that it works in the copyright holder's environment, we cannot guarantee that
# there will be no defects.
#
# If you modify the source code, do not remove the copyright notice.
#
# Copyright holder shall not be liable for any claim, damages or other liability arising
# from or relating to the source code or arising from the use or other dealings in the software.

"""
Silkey Camera(event-base) View Started.
"""

import global_value as g_value
import global_define as g_define
from silky_util import CameraConfig, BiasConfig

from metavision_core.event_io.raw_reader import initiate_device
from metavision_core.event_io import EventsIterator
from metavision_sdk_core import OnDemandFrameGenerationAlgorithm, ColorPalette, FlipXAlgorithm
from metavision_hal import I_TriggerIn
from metavision_sdk_ui import EventLoop
from vmbpy import Log

import copy
import numpy as np
import os
import configparser

# 'config.ini' file value getting
config = configparser.ConfigParser()
config.read('./config/config.ini',encoding="utf-8")
common = 'common'
FRAME_DELAY_NUM = int(config.get(common, 'frame_delay_num'))


def main():

    """ Main """
    args = g_define.parse_args()
    log = Log.get_instance()

    # first we create a HAL device
    try:
        device = initiate_device(path=args.event_camera_serial)
    except(OSError):
        log.error('[Error] Event camera not found. Check the USB connection.')
        g_value.event_camera_err_flag.set()

    g_value.event_camera_open_proc.set()
    if g_value.event_camera_err_flag.is_set():
        return

    # Wait for the end of the 'frame camera' opening
    g_value.frame_camera_open_proc.wait()
    if g_value.frame_camera_open_proc.is_set():
        if g_value.frame_camera_err_flag.is_set():
            return # unavoidably terminated

    # Events iterator on Camera
    frame_period_us = int(1e6/30)  # fps
    mv_iterator = EventsIterator.from_device(device=device, delta_t=frame_period_us, relative_timestamps=False)
    height, width = mv_iterator.get_size()  # Camera Geometry

    # Camera Setting
    cc = CameraConfig(device)
    cc.set_from_file(file_name='./config/settings.json', disp_flag=False)

    # Biases Setting
    if args.using_bias_file:
        biases = mv_iterator.reader.device.get_i_ll_biases()
        bc = BiasConfig(biases)
        bc.set_from_file(file_name='./config/biases.bias')

    # Event Frame Generator
    on_demand_gen = OnDemandFrameGenerationAlgorithm(width, height, args.accumulation_time_us)
    on_demand_gen.set_color_palette(ColorPalette.Dark)

    # Reverse filter
    flipX_filter = FlipXAlgorithm(width - 1)
    event_buf = flipX_filter.get_empty_output_buffer()

    # Process events
    next_processing_ts = frame_period_us
    frame = np.zeros((height, width, 3), np.uint8)

    # For slip out of frame
    g_value.b_frame_list = list([0] * FRAME_DELAY_NUM)

    g_value.is_recording = False
    g_value.is_as_seen_recording = False
    g_value.is_evt_recording = False
    init_recording = False

    for evs in mv_iterator:
        EventLoop.poll_and_dispatch()  # Dispatch system events to the window

        if len(evs) == 0:
            continue

        # Reverse
        if args.disable_reverse:
            event_buf = evs
        else:
            # Reverse Mode
            flipX_filter.process_events(evs, event_buf)

        # Process events
        on_demand_gen.process_events(event_buf)

        # Trigger new frame generations as long as the last event is high enough
        ts = evs["t"][-1]
        while(ts > next_processing_ts):
            on_demand_gen.generate(next_processing_ts, frame)
            next_processing_ts += frame_period_us

            # For slip out of frame
            if len(g_value.b_frame_list) > 1:
                g_value.b_frame_list.pop(0)
                g_value.b_frame_list.append(copy.deepcopy(frame))
            else:
                g_value.b_frame_list[0] = copy.deepcopy(frame)

        # Catch main end event
        if g_value.is_alive == False:
            break

        # Catch main recording event
        if g_value.is_evt_recording == True:
            if not init_recording:

                # InputTrigger Setting
                i_trigger_in = device.get_i_trigger_in()
                i_trigger_in.enable(I_TriggerIn.Channel.MAIN)

                log_path = g_value.recording_name + '.raw'
                log_path = os.path.join('.', 'video', log_path)
                log.info('Event-Raw Recording to {} Start...'.format(log_path))
                device.get_i_events_stream().log_raw_data(log_path)
            init_recording = True
        else:
            if init_recording:
                # InputTrigger Setting
                i_trigger_in = device.get_i_trigger_in()
                i_trigger_in.disable(I_TriggerIn.Channel.MAIN)

                log.info('Event-Raw Recording Stop...')
                device.get_i_events_stream().stop_log_raw_data()
            init_recording = False

    del device
    log.info('Thread \'Event-Camera\' terminated.')


if __name__ == '__main__':
    main()
