# Copyright (c) 2024 CenturyArks Co.,Ltd.
#
# These source code may be used (including the use of modified source code) only when using
# CenturyArks's products or services.
#
# Although we have confirmed that it works in the copyright holder's environment, we cannot guarantee that
# there will be no defects.
#
# If you modify the source code, do not remove the copyright notice.
#
# Copyright holder shall not be liable for any claim, damages or other liability arising
# from or relating to the source code or arising from the use or other dealings in the software.

"""
Alvium Camera(global shutter) View Started.
"""

import bothview_event as silky_cam
import global_value as g_value
import global_define as g_define
import global_calc as g_calc

import threading
import sys
import cv2
from typing import Optional
from vmbpy import *
from metavision_sdk_ui import EventLoop, BaseWindow, MTWindow, UIKeyEvent

import configparser
import queue
import numpy
import datetime, time, math, ffmpeg, os
import copy
import math
import re
from decimal import *
import subprocess

# 'config.ini' file value getting
config = configparser.ConfigParser()
config.read('./config/config.ini',encoding="utf-8")
conf_camera01 = 'camera01_spec'
conf_camera02 = 'camera02_spec'
common = 'common'
user_setting = 'user_setting'
FRAME_RESOLUTION_W = int(config.get(conf_camera01, 'resolution_w'))
FRAME_RESOLUTION_H = int(config.get(conf_camera01, 'resolution_h'))
EVENT_RESOLUTION_W = int(config.get(conf_camera02, 'resolution_w'))
EVENT_RESOLUTION_H = int(config.get(conf_camera02, 'resolution_h'))
FRAME_PIXCEL_W = float(config.get(conf_camera01, 'pixcel_w'))
FRAME_PIXCEL_H = float(config.get(conf_camera01, 'pixcel_h'))
EVENT_PIXCEL_W = float(config.get(conf_camera02, 'pixcel_w'))
EVENT_PIXCEL_H = float(config.get(conf_camera02, 'pixcel_h'))
FRAME_QUEUE_SIZE = int(config.get(common, 'frame_queue_size'))
ADJUST_VIEW_W = int(config.get(user_setting, 'adjust_view_w'))
ADJUST_VIEW_H = int(config.get(user_setting, 'adjust_view_h'))

g_value.FFMPEG_PROCESS = None
g_value.RECORDING_FRAME_CNT = 0

# Multi-view mode: Explanatory image
desc_img1 = cv2.imread('images/description.jpg')
desc_img2 = cv2.imread('images/multi_frame_discription.jpg')
desc_list = [desc_img1, desc_img2]


def print_preamble():
    print('///////////////////////////////////////////////')
    print('// SilkyEvCam - BothView                  /////')
    print('///////////////////////////////////////////////')
    print('Press <Esc or Q> to exit. ')
    print('Press <V> to Change View Mode.')
    print('Press <R> to Recording(2video mode) Start or End. ')
    print('Press <M> to Recording(multi-view mode) Start or End. ')
    print(flush=True)


def abort(reason: str, return_code: int = 1):
    print(reason + '\n')
    sys.exit(return_code)


def concat_tile(im_list_2d):
    return cv2.vconcat([cv2.hconcat(im_list_h) for im_list_h in im_list_2d])


def change_from_frame_to_opencv(frame: Frame, id_no: int = -1) -> numpy.ndarray:
    cv_frame = frame.as_opencv_image()

    if id_no != -1:
        cv2.putText(cv_frame, str(id_no), org=(g_value.write_frame_id_x, g_value.write_frame_id_y), 
            fontScale=2, color=(0,0,255), thickness=1, fontFace=cv2.FONT_HERSHEY_COMPLEX_SMALL, lineType=cv2.LINE_AA)

    return cv_frame


def try_put_frame(q: queue.Queue, cam: Camera, frame: Optional[Frame]):
    # Add frame information of the target camera (id) to the queue.
    try:
        q.put_nowait((cam.get_id(), frame))

    except queue.Full:
        pass


def createMp4Process(img_w: int, img_h: int, bitrate=-1) -> subprocess.Popen:
    # Generate process during recording
    log = Log.get_instance()
    args = g_define.parse_args()

    # Constants
    _CODEC_DICT = {'GPU':'h264_nvenc', 'CPU':'h264'} # No used 'AMF':'h264_amf', 'QSV':'h264_qsv'

    v_codec = _CODEC_DICT['GPU'] if args.frame_using_gpu else _CODEC_DICT['CPU']
    if bitrate==-1:
        bitrate = args.frame_bitrate

    file_path = g_value.recording_name + '.mp4'
    file_path = os.path.join('.', "video", file_path)

    # Memo:
    # To minimize missing frames, "vsync=passthrough" is used.
    # Normally set "vsync=vfr".
    ffmpeg_process = (
        ffmpeg
        .input('pipe:', format='rawvideo', pix_fmt='bgr24',
            s='{}x{}'.format(img_w, img_h), use_wallclock_as_timestamps=1)
        .crop(g_value.img_trim_offset_x, g_value.img_trim_offset_y, g_value.img_trim_width, g_value.img_trim_height)
        .output(file_path, pix_fmt='yuv420p', vcodec=v_codec, vsync='passthrough', video_bitrate='{}k'.format(bitrate), loglevel="info")
        .run_async(pipe_stdin=True, overwrite_output=True) # Display log on prompt
    )

    log.info('Frame-MP4 Recording to {} Start...'.format(file_path))
    return ffmpeg_process


def destroyMp4Process(ffmpeg_process: subprocess.Popen):
    # Destroys the process being recorded
    log = Log.get_instance()

    if ffmpeg_process != None:
        log.info('Frame-MP4 Recording Stop...')
        ffmpeg_process.stdin.close()
        ffmpeg_process.wait()
        ffmpeg_process = None


# Thread Objects
class FrameProducer(threading.Thread):
    def __init__(self, cam: Camera, frame_queue: queue.Queue):
        threading.Thread.__init__(self)

        self.log = Log.get_instance()
        self.cam = cam
        self.frame_queue = frame_queue
        self.killswitch = threading.Event()

        # Instance Variables for Recordings
        self.init_recording = False
        self.ffmpeg_process = None
        self.recording_frame_cnt = 0
        self.args = g_define.parse_args()

    def __call__(self, cam: Camera, stream: Stream, frame: Frame):
        # This method is executed within VmbC context. All incoming frames
        # are reused for later frame acquisition. If a frame shall be queued, the
        # frame must be copied and the copy must be sent, otherwise the acquired
        # frame will be overridden as soon as the frame is reused.
        if frame.get_status() == FrameStatus.Complete:

            # Display frame number
            #print('{} acquired {}'.format(cam, frame), flush=True)

            frame_cpy = copy.deepcopy(frame)

            if g_value.is_recording == True:

                frame_id = frame.get_id() if self.args.write_frame_id else -1
                cv_frame = change_from_frame_to_opencv(frame_cpy, frame_id)
                try:
                    g_value.FFMPEG_PROCESS.stdin.write(cv_frame)
                    g_value.RECORDING_FRAME_CNT += 1
                except Exception as e:
                    # Select 'GPU' and follow up if GPU is not available.
                    self.log.error('ffmpeg error:' + str(e))
                    g_value.is_recording = False
                    g_value.is_evt_recording = False

            if not self.frame_queue.full():
                try_put_frame(self.frame_queue, cam, frame_cpy)

        else:
            # Memo: 
            # If this point is reached during recording, 
            # the mp4 file will have no frames.
            # This also happens when outputting a file 
            # with "FFMPEG_PROCESS". (The log output will show "drop=n").
            #
            # The cause depends mainly on the number of times the FPS value is processed 
            # and computer specs, so adjust accordingly.
            self.log.warning('Missing frame... frame info:{}'.format(frame))


        cam.queue_frame(frame)

    def stop(self):
        self.killswitch.set()

    def setup_camera(self):

        args = g_define.parse_args()

        # 1. Initialize 'ROI-related' settings
        self.cam.BinningHorizontal.set(1)
        self.cam.BinningVertical.set(1)
        width_max = self.cam.WidthMax.get()
        height_max = self.cam.HeightMax.get()
        self.cam.OffsetX.set(0)
        self.cam.OffsetY.set(0)
        self.cam.Width.set(width_max)
        self.cam.Height.set(height_max)

        # 2.Sensor size Calculation START
        frame_sensor_w, frame_sensor_h = g_calc.get_cencer_size((FRAME_RESOLUTION_W, FRAME_RESOLUTION_H), 
                                                                (FRAME_PIXCEL_W, FRAME_PIXCEL_H))

        event_sensor_w, event_sensor_h = g_calc.get_cencer_size((EVENT_RESOLUTION_W, EVENT_RESOLUTION_H), 
                                                                (EVENT_PIXCEL_W, EVENT_PIXCEL_H))

        # 3.Image Resolution size Calculation START
        frame_trim_w, frame_trim_h = g_calc.get_trim_pixel_size((FRAME_RESOLUTION_W, FRAME_RESOLUTION_H), 
                                                                (EVENT_RESOLUTION_W, EVENT_RESOLUTION_H), 
                                                                (frame_sensor_w, frame_sensor_h),
                                                                (event_sensor_w, event_sensor_h))

        # 4.Frame size calculation for ROI specification
        frame_roi_w = FRAME_RESOLUTION_W - (frame_trim_w * 2)
        frame_roi_h = FRAME_RESOLUTION_H - (frame_trim_h * 2)

        # 5.Adjust ROI to meet frame camera specifications
        ((frame_roi_w, frame_roi_h), (frame_trim_w, frame_trim_h))= g_calc.get_adjusted_roi((frame_roi_w, frame_roi_h), 
                                                                                            (frame_trim_w, frame_trim_h))

        # 6.Obtain ROI information (for image cropping)
        # * Because the Frame ROI Interval is large in 4px increments, Crop the image itself.

        # Frame Position Adjustment
        adjust_w, adjust_h = ADJUST_VIEW_W, ADJUST_VIEW_H
        if args.disable_adjust:
            adjust_w, adjust_h = 0, 0

        g_value.img_trim_width = int(frame_roi_w)
        g_value.img_trim_height = int(frame_roi_h)
        g_value.img_trim_offset_x = int(frame_trim_w - g_calc.get_adjusted_offset(adjust_w))
        g_value.img_trim_offset_y = int(frame_trim_h + g_calc.get_adjusted_offset(adjust_h))
        g_value.write_frame_id_x = 40 + g_value.img_trim_offset_x
        g_value.write_frame_id_y = 60 + g_value.img_trim_offset_y

        # 7.Initialize output trigger settings
        self.cam.LineMode.set('Output')
        self.cam.LineSource.set('Off')
        self.cam.TriggerSource.set('Line0')
        self.cam.LineInverter.set(False)

        '''
        # Try to enable automatic exposure time setting
        try:
            self.cam.ExposureAuto.set('Once')

        #except (AttributeError, VmbFeatureError):
            self.log.info('Camera {}: Failed to set Feature \'ExposureAuto\'.'.format(
                        self.cam.get_id()))
        '''

        # 8.Basic color. Monochrome if not obtained.
        cv_fmts = intersect_pixel_formats(self.cam.get_pixel_formats(), OPENCV_PIXEL_FORMATS)
        color_fmts = intersect_pixel_formats(cv_fmts, COLOR_PIXEL_FORMATS)

        if color_fmts:
            self.cam.set_pixel_format(color_fmts[0])

        else:
            mono_fmts = intersect_pixel_formats(cv_fmts, MONO_PIXEL_FORMATS)

            if mono_fmts:
                self.cam.set_pixel_format(mono_fmts[0])

            else:
                abort('Camera does not support a OpenCV compatible format natively. Abort.')

        # 9.Enable Acquisition FrameRate Control
        try:
            if args.eaf_mode:
                self.log.info('--> [setup_camera] AcquisitionFrameRateEnable (value: %sfps)' % args.acquisition_frame)
                self.cam.AcquisitionFrameRateEnable.set(1) # Mode On
                self.cam.AcquisitionFrameRate.set(args.acquisition_frame) # FPS
        
        except VmbFeatureError as e:
            self.log.warning('\'AcquisitionFrameRate\' is not set.')
            pass

    # Reads the camera's configuration XML file.
    def load_setting_camera(self):
        self.log.info('--> [load_setting_camera] Frame Camera has been opened ({}[{}])'.format(self.cam.get_serial(), self.cam.get_id()))

        settings_file = 'config/settings.xml'

        # Load camera settings from file.
        self.cam.load_settings(settings_file, PersistType.All)

        self.log.info('--> [load_setting_camera] Feature values have been loaded from given file \'%s\'' % settings_file)

    def run(self):
        self.log.info('Thread \'FrameProducer({}[{}])\' started.'.format(self.cam.get_serial(), self.cam.get_id()))

        try:
            with self.cam:
                self.load_setting_camera()
                self.setup_camera()

                try:
                    self.cam.start_streaming(self)
                    self.killswitch.wait()

                finally:
                    self.cam.stop_streaming()

        except VmbCameraError:
            pass

        finally:
            try_put_frame(self.frame_queue, self.cam, None)

        self.log.info('Thread \'FrameProducer({}[{}])\' terminated.'.format(self.cam.get_serial(), self.cam.get_id()))


class FrameConsumer(threading.Thread):
    # Performs main "video compositing" processing

    # Constants
    _VIEW_DICT = {'multi':0, 'both':1, 'event':2, 'frame':3}
    _IMAGE_CAPTION = 'SilkyEvCam - BothView'

    def __init__(self, cam: Camera, frame_queue: queue.Queue):
        threading.Thread.__init__(self)

        self.log = Log.get_instance()
        self.cam = cam
        self.frame_queue = frame_queue
        self.window = None

        # Variables for propagation to 'callback functions'
        self.quit_mode = 0
        self.view_mode = 0
        self.record_mode = 0
        self.as_seen_record_mode = 0

        # For multi-view mode only: Explanatory image
        self.init_multi_view = False # Multi-mode initial display
        self.desc_img1 = cv2.imread('images/description.jpg')
        self.desc_img2 = cv2.imread('images/multi_frame_discription.jpg')
        self.desc_list = [self.desc_img1, self.desc_img2]

        # Instance Variables for Recordings
        self.init_recording = False
        self.ffmpeg_process = None

    # Stopping the camera
    def closeCamera(self):
        self.log.info('Window close process...')
        g_value.is_alive = False
        if self.window:
            self.cam.stop_streaming()
            self.window.set_close_flag()
            time.sleep(1) # Wait
            self.window.destroy()

    def execRecResultAnalysis(self):
        g_value.is_rec_result_proc = True
        self.log.info('Start analyzing file...')
        file_path = os.path.join('.', "video", g_value.recording_name)
        f_total = self.getResultFrameNum(file_path + '.mp4')
        e_total = self.getResultEventNum(file_path + '.raw')
        log_list = ['Event Ext Triggers: ' + str(e_total),
                    'Frame Actual Count: ' + str(g_value.RECORDING_FRAME_CNT),
                    'Frame Result Count: ' + str(f_total),
                    'Frame Diff   Count: ' + str(g_value.RECORDING_FRAME_CNT - f_total)]

        result_path = file_path + '_result.txt'
        with open(result_path, mode='w') as f:
            f.writelines([x + '\n' for x in log_list])
            f.close()
        self.log.info('Finish analyzing file... The text of the recording results is output to \'{}\''.format(result_path))
        g_value.is_rec_result_proc = False

    def getResultFrameNum(self, file_path: str):

        cap=cv2.VideoCapture(file_path)
        total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))

        return total

    def getResultEventNum(self, file_path: str):

        cmd = 'metavision_file_info -i '
        regix = r'External triggers(\s+)([0-9]*)(\s+)'
        exec = cmd + file_path
        result = 0

        try:
            proc = subprocess.run(exec, shell=True,
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            
            proc_result = proc.stderr.decode('shift_jis') # Possible inclusion of sjis-code
            match = re.search(regix, proc_result)
            if match != None:
                group = match.group()
                result = int(group.split()[2])

        except subprocess.CalledProcessError:
            self.log.error('Failed to execute external program ({})'.format(cmd))

        return result

    def run(self):

        frames = {}
        alive = True

        self.log.info('Thread \'FrameConsumer\' started.')

        # keyboard callback (Internal Functions)
        def keyboard_cb(key, scancode, action, mods):
            global g_value

            if (action == 1): # Response to "onkeydown:action=1" and "onkeyup:action=0" being triggered twice.
                if key == UIKeyEvent.KEY_ESCAPE or key == UIKeyEvent.KEY_Q:
                    self.quit_mode = 1
                    self.log.info('OnKey Viewer End')

                elif key == UIKeyEvent.KEY_V:
                    if self.record_mode == 1 or self.as_seen_record_mode == 1:
                        return

                    self.view_mode = (self.view_mode + 1) % 4 # Switching between 0 and 3
                    msg = [k for k, v in self._VIEW_DICT.items() if v == self.view_mode][0]
                    self.log.info('OnKey View Mode: ' + msg)

                elif key == UIKeyEvent.KEY_R:
                    if self.as_seen_record_mode == 1:
                        return

                    self.record_mode = (self.record_mode + 1) % 2 # Switching between 0 and 1
                    if self.record_mode == 1:
                        g_value.recording_name = 'recording_' + datetime.datetime.now().strftime('%y%m%d_%H%M%S_%f')[:-3]
                        msg = 'On'

                        # Memo:
                        # Output trigger is output when "Invert" is set.
                        # It should not be included in the event trigger (raw) record.
                        self.cam.stop_streaming()
                        self.cam.LineMode.set('Output')
                        self.cam.LineSource.set('ExposureActive')
                        self.cam.TriggerSource.set('Line0')
                        self.cam.LineInverter.set(True)

                        # Camera Frame Size getting
                        width_val = self.cam.Width.get()
                        height_val = self.cam.Height.get()

                        # Start event RAW
                        g_value.is_evt_recording = True

                        # Start frame MP4
                        self.log.info('create_mp4_process-start:' + datetime.datetime.now().strftime('%y%m%d_%H%M%S_%f')[:-3])
                        g_value.FFMPEG_PROCESS = createMp4Process(width_val, height_val)
                        self.log.info('create_mp4_process-end:' + datetime.datetime.now().strftime('%y%m%d_%H%M%S_%f')[:-3])
                        time.sleep(1) # Wait

                        g_value.is_recording = True
                        g_value.RECORDING_FRAME_CNT = 0

                        # Memo:
                        # The recording process in GPU mode rarely freezes.(ffmpeg)
                        # To prevent missing frames in such cases, increase the buffer size.
                        self.cam.start_streaming(FrameProducer(self.cam, self.frame_queue), buffer_count=50)

                    else:
                        msg = 'Off'

                        self.cam.stop_streaming()
                        self.cam.LineSource.set('Off')
                        g_value.is_recording = False
                        
                        destroyMp4Process(g_value.FFMPEG_PROCESS)

                        # Stop event RAW
                        time.sleep(1) # Wait
                        g_value.is_evt_recording = False

                        self.cam.start_streaming(FrameProducer(self.cam, self.frame_queue))

                        thd = threading.Thread(target=self.execRecResultAnalysis)
                        thd.start()

                    self.log.info('OnKey Record(2video) Mode: ' + msg)
                    self.log.info('During recording, the view is fixed to "frame mode"...')
                    self.view_mode = self._VIEW_DICT['frame']

                elif key == UIKeyEvent.KEY_M:
                    if self.record_mode == 1:
                        return

                    self.as_seen_record_mode = (self.as_seen_record_mode + 1) % 2 # Switching between 0 and 1

                    if self.as_seen_record_mode == 1:
                        g_value.recording_name = 'multi_recording_' + datetime.datetime.now().strftime('%y%m%d_%H%M%S_%f')[:-3]
                        msg = 'On'
                        g_value.is_as_seen_recording = True
                    else:
                        msg = 'Off'
                        g_value.is_as_seen_recording = False

                    self.log.info('OnKey Record(As seen) Mode: ' + msg)
                    self.view_mode = self._VIEW_DICT['multi']


        try:
            while alive:

                # Dispatch system events to the window
                # Similar to "cv2.waitKey(1)"
                EventLoop.poll_and_dispatch() 

                # Update current state by dequeuing all currently available frames.
                frames_left = self.frame_queue.qsize()
                while frames_left:
                    try:
                        cam_id, frame = self.frame_queue.get_nowait()

                    except queue.Empty:
                        break

                    # Add/Remove frame from current state.
                    if frame:
                        frames[cam_id] = frame

                    else:
                        frames.pop(cam_id, None)

                    frames_left -= 1

                # Construct image by stitching frames together.
                if frames:

                    cv_images = [change_from_frame_to_opencv(frames[cam_id]) for cam_id in sorted(frames.keys())]

                    # Overlay images.
                    img_frame = numpy.concatenate(cv_images, axis=1)
                    img_event = g_value.b_frame_list[0]

                    # Crop from frame image. (ROI)
                    img_frame = img_frame[g_value.img_trim_offset_y:(g_value.img_trim_offset_y + g_value.img_trim_height), 
                                        g_value.img_trim_offset_x:(g_value.img_trim_offset_x + g_value.img_trim_width)]

                    # Resizes the event side (smaller image) to the size of the frame side (larger image)
                    f_h, f_w, _= numpy.shape(img_frame)
                    img_event = cv2.resize(img_event,(f_w, f_h))

                    img_display = self.getImageForViewMode(img_frame, img_event)

                    # Execute 'As seen' REC 
                    self.selectRecordMode(img_display)

                    # '2video' REC
                    if g_value.is_recording:
                        description = '[REC(2video)]'
                        color = (0,0,255) # R
                        cv2.putText(img_display, description, org=(30, 30), fontScale=1, color=color, 
                            thickness=1, fontFace=cv2.FONT_HERSHEY_COMPLEX_SMALL, lineType=cv2.LINE_AA)

                    # 'As seen' REC
                    elif g_value.is_as_seen_recording:

                        description = '[REC(multi)]'
                        color = (0,0,255) # R
                        cv2.putText(img_display, description, org=(30, 30), fontScale=1, color=color, 
                            thickness=1, fontFace=cv2.FONT_HERSHEY_COMPLEX_SMALL, lineType=cv2.LINE_AA)
                        
                    else:
                        try:
                            if g_value.is_rec_result_proc:
                                # '2video' REC only
                                description = 'Analyzing results. Please wait...'
                                color = (0,255,0) # G
                                cv2.putText(img_display, description, org=(30, 30), fontScale=1, color=color, 
                                    thickness=1, fontFace=cv2.FONT_HERSHEY_COMPLEX_SMALL, lineType=cv2.LINE_AA)
                        except AttributeError:
                            g_value.is_rec_result_proc = False

                        self.record_mode = 0

                    # Metavision Window
                    if self.window == None:
                        # Initial window display size
                        self.window = MTWindow(title=self._IMAGE_CAPTION, width=f_w, height=f_h, mode=BaseWindow.RenderMode.BGR, open_directly=True)
                        # Callbacks for various events
                        self.window.set_keyboard_callback(keyboard_cb)

                    if not self.window.should_close() and self.quit_mode == 0:
                        self.window.show_async(img_display)
                    else:
                        if g_value.is_recording:
                            self.log.info('It was being recorded(2video). Recording will be terminated...')
                            keyboard_cb(key=UIKeyEvent.KEY_R, action=1, scancode=None, mods=None) # R OFF
                            time.sleep(1) # Wait
                        elif g_value.is_as_seen_recording:
                            self.log.info('It was being recorded(multi). Recording will be terminated...')
                            keyboard_cb(key=UIKeyEvent.KEY_M, action=1, scancode=None, mods=None) # M OFF
                            time.sleep(1) # Wait

                        break

        finally:
            # Normal termination in case of unforeseen errors.
            self.closeCamera()
            self.log.info('Thread \'FrameConsumer\' terminated.')

    def selectRecordMode(self, img_frame: numpy.ndarray):

        # Select record mode
        if g_value.is_as_seen_recording:
            if not self.init_recording:

                # 1.Start Recording
                self.log.info('create_mp4_process-start:' + datetime.datetime.now().strftime('%y%m%d_%H%M%S_%f')[:-3])
                img_h, img_w, _ = numpy.shape(img_frame)

                # Specifies the bit rate value at which 'block noise' is not generated.
                self.ffmpeg_process = createMp4Process(img_w, img_h, 4000)
                self.log.info('create_mp4_process-end:' + datetime.datetime.now().strftime('%y%m%d_%H%M%S_%f')[:-3])

                self.init_recording = True
            else:
                try:
                    self.ffmpeg_process.stdin.write(img_frame)
                except Exception as e:
                    # Select 'GPU' and follow up if GPU is not available.
                    self.log.error('ffmpeg error:' + str(e))
                    g_value.is_as_seen_recording = False
                    self.as_seen_record_mode = 0
                    self.init_recording = False

        else:
            if self.init_recording:

                try:
                    # Save the last frame
                    self.ffmpeg_process.stdin.write(img_frame)
                except Exception as e:
                    # Select 'GPU' and follow up if GPU is not available.
                    self.log.error('ffmpeg error:' + str(e))

                # 2.End of Recording
                destroyMp4Process(self.ffmpeg_process)

            self.init_recording = False

    def getImageForViewMode(self, img_frame: numpy.ndarray, img_event: numpy.ndarray):

        # Composite image (Internal Functions)
        def doImageComposition(img_frame: numpy.ndarray, img_event: numpy.ndarray):
            # alpha and beta are added together, and the larger number is used rather than 1.0 because it was easier to see.
            return cv2.addWeighted(
                src1=img_frame, alpha=0.8,
                src2=img_event, beta= 0.8,
                gamma=0
            )
        
        # Create split image (Internal Functions)
        def doImageConcatenation(im_list_2d: list):
            return cv2.vconcat([cv2.hconcat(im_list_h) for im_list_h in im_list_2d])

        if self.view_mode == self._VIEW_DICT['both']:
            img_display = doImageComposition(img_frame, img_event)
        elif self.view_mode == self._VIEW_DICT['event']:
            img_display = img_event
        elif self.view_mode == self._VIEW_DICT['frame']:
            img_display = img_frame
        elif self.view_mode == self._VIEW_DICT['multi']:
            img_display = doImageComposition(img_frame, img_event)
            img_display_s = cv2.resize(img_display, dsize=(0, 0), fx=0.5, fy=0.5)
            img_event_s = cv2.resize(img_event, dsize=(0, 0), fx=0.5, fy=0.5)
            img_frame_s = cv2.resize(img_frame, dsize=(0, 0), fx=0.5, fy=0.5)

            if not self.init_multi_view:
                re_h, re_w, _  = numpy.shape(img_display_s)
                self.desc_list[0] = cv2.resize(self.desc_list[0], dsize=(re_w, re_h))
                self.desc_list[1] = cv2.resize(self.desc_list[1], dsize=(re_w, re_h))
                self.init_multi_view = True

            dt = datetime.datetime.now()
            sec_2nd_digit = math.floor(dt.second / 10)
            img_description = self.desc_list[0] if sec_2nd_digit % 2 == 0 else self.desc_list[1]

            img_display = doImageConcatenation([[img_display_s, img_frame_s],[img_event_s, img_description]])
            del img_display_s
            del img_event_s
            del img_frame_s
            del img_description

        return img_display


class MainThread(threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self)

        self.frame_queue = queue.Queue(maxsize=FRAME_QUEUE_SIZE) 
        self.producers = {}
        self.producers_lock = threading.Lock() # Creating Lock Objects
        self.args = g_define.parse_args()

    def __call__(self, cam: Camera, event: CameraEvent):
        if event == CameraEvent.Detected:
            with self.producers_lock:
                self.producers[cam.get_id()] = FrameProducer(cam, self.frame_queue)
                self.producers[cam.get_id()].start()

        elif event == CameraEvent.Missing:
            with self.producers_lock:
                producer = self.producers.pop(cam.get_id())
                producer.stop()
                producer.join()

    def run(self):
        log = Log.get_instance()

        consumer = None
        vmb = VmbSystem.get_instance()
        vmb.enable_log(LOG_CONFIG_INFO_CONSOLE_ONLY)

        log.info('Thread \'MainThread\' started.')

        with vmb:
            cam = None
            try:
                cams = vmb.get_all_cameras()
                cam = cams[0]

                if self.args.frame_camera_serial != '':
                    target_flag = False
                    for c in cams:
                        if c.get_serial() == self.args.frame_camera_serial:
                            target_flag = True
                            cam = c
                            break
                    if not target_flag:
                        raise IndexError("No frame camera for the specified serial value of the parameter.")
            except(IndexError):
                log.error('[Error] Frame camera not found. Check the USB connection.')
                g_value.frame_camera_err_flag.set()

            g_value.frame_camera_open_proc.set()
            if g_value.frame_camera_err_flag.is_set():
                return

            # Wait for the end of the 'event camera' opening
            g_value.event_camera_open_proc.wait()
            if g_value.event_camera_open_proc.is_set():
                if g_value.event_camera_err_flag.is_set():
                    return # unavoidably terminated

            self.producers[cam.get_id()] = FrameProducer(cam, self.frame_queue)
            consumer = FrameConsumer(cam, self.frame_queue)

            with self.producers_lock:
                for producer in self.producers.values():
                    producer.start()

            # Start and wait for consumer to terminate
            vmb.register_camera_change_handler(self)
            consumer.start()
            consumer.join()
            vmb.unregister_camera_change_handler(self)

            # Stop all FrameProducer threads
            with self.producers_lock:
                # Initiate concurrent shutdown
                for producer in self.producers.values():
                    producer.stop()

                # Wait for shutdown to complete
                for producer in self.producers.values():
                    producer.join()

        log.info('Thread \'MainThread\' terminated.')


if __name__ == '__main__':

    # For help display
    g_define.parse_args()

    print_preamble()
    g_value.is_alive = True
    g_value.frame_camera_open_proc = threading.Event()
    g_value.frame_camera_err_flag = threading.Event()
    g_value.event_camera_open_proc = threading.Event()
    g_value.event_camera_err_flag = threading.Event()

    frame_thread = MainThread()
    event_thread = threading.Thread(target=silky_cam.main)

    # Thread Start
    frame_thread.start()
    event_thread.start()

    # Thread Wait
    frame_thread.join()
    event_thread.join()
    g_value.is_alive = False
